from commonutils import StaticUtils
from struct import calcsize, pack_into, unpack_from

class Field:
   __encoding = "utf-8"
   
   __formats = {
      1: ["B"],
      2: ["H"],
      3: ["L1"],
      4: ["L"],
      5: ["Q3"],
      6: ["Q2"],
      7: ["Q1"],
      8: ["Q"]
   }
   
   for v in __formats.values():
      v.append(v[0].lower())
   
   def __init__(self, size = None, **kw):
      previousField = kw.get("previousField")
      signed = kw.get("signed")
      value = kw.get("value")
      valueIsStr = not size
      
      self.__offset = previousField.nextOffset if previousField else 0
      self.__size = len(value) if valueIsStr else size
      
      self.__format = f"{self.__size}s" if valueIsStr else None if signed == None else Field.__formats[self.__size][+signed]
      
      self.__value = value.encode(Field.__encoding) if valueIsStr else value
   
   @property
   def nextOffset(self):
      return self.__offset + self.__size
   
   @property
   def offset(self):
      return self.__offset
   
   @offset.setter
   def offset(self, offset):
      self.__offset = offset
   
   @property
   def size(self):
      return self.__size
   
   @property
   def value(self):
      return self.__value
   
   @value.setter
   def value(self, value):
      if self.__format != None:
         self.__value = value
      
      elif len(value) != self.__size:
         raise ValueError()
      
      else:
         self.__value = value[:]
   
   @staticmethod
   def createChain(size = None, signed = None, value = None, fields = None):
      chain = fields
      
      if chain:
         for i, field in enumerate(chain):
            if i:
               field.__offset = chain[i - 1].nextOffset
      
      else:
         chain = []
         
         args = (size, signed, value)
         
         def count():
            for x in args:
               if StaticUtils.isIterable(x):
                  return len(x)
            
            raise ValueError()
         
         for index in range(count()):
            params = tuple(v[index] if StaticUtils.isIterable(v) else v for v in args)
            
            chain.append(Field(
               params[0],
               previousField = chain[index - 1] if index else None,
               signed = params[1],
               value = params[2]))
      
      return chain
   
   def _get(self, buf, offset, byteorder):
      i = offset + self.__offset
      j = i + self.__size
      
      if self.__format == None:
         self.__value = buf[i:j]
      
      else:
         isStr = isinstance(self.__format[0], int)
         
         if len(self.__format) != 2 or isStr:
            self.__value = unpack_from(f"{byteorder}{self.__format}", buf, i)[0]
            
            if isStr:
               self.__value = self.__value.decode(Field.__encoding)
         
         else:
            self.__value = buf[i:j]
            
            for _ in range(int(self.__format[1])):
               if byteorder == ">":
                  self.__value[:0] = [0]
               
               else:
                  self.__value.append(0)
            
            self.__value = unpack_from(f"{byteorder}{self.__format[0]}", self.__value, 0)[0]
      
      return self
   
   def _set(self, buf, offset, byteorder):
      i = offset + self.__offset
      j = i + self.__size
      
      if self.__format == None:
         buf[i:j] = self.__value
      
      elif len(self.__format) != 2 or isinstance(self.__format[0], int):
         pack_into(f"{byteorder}{self.__format}", buf, i, self.__value)
      
      else:
         fmt = f"{byteorder}{self.__format[0]}"
         tmpbuf = bytearray(calcsize(fmt))
         
         pack_into(fmt, tmpbuf, 0, self.__value)
         
         for _ in range(int(self.__format[1])):
            tmpbuf.pop(0 if byteorder == ">" else -1)
         
         buf[i:j] = tmpbuf



from .field import Field

class Packet:
   def __init__(self, format, **kwargs):
      self.__format = format
      
      if "cmd" in kwargs:
         sizeSpecified = "size" in kwargs
         paramsSpecified = "params" in kwargs
         
         if sizeSpecified and paramsSpecified:
            raise ValueError("Either 'params' or 'size' can be specified.")
         
         self.__buf = bytearray([0] * format.minPacketSize)
         
         if sizeSpecified:
            self.__buf.extend(bytearray([0] * kwargs["size"]))
         
         elif paramsSpecified:
            self.__buf[format._paramsOffset : format._paramsOffset] = kwargs["params"]
         
         format.setCommandNumber(self.__buf, kwargs["cmd"])
         format.finalizePacket(self.__buf)
   
   def wrap(self, buffer, **kwargs):
      if not isinstance(buffer, bytearray):
         buffer = bytearray(buffer)
      
      self.__buf = buffer[kwargs.get("start", 0):kwargs.get("end", len(buffer))]

      if not self.__format.isValid(self.__buf):
         raise ValueError(f"{self.__buf} is not a valid packet.")
      
      return self
   
   def getParam(self, param):
      return self.__format.getParam(self.__buf, param)
   
   def setParam(self, param, finalize = False):
      self.__format.setParam(self.__buf, param, finalize)
   
   def setParams(self, values = None, signed = None, size = None, fields = None):
      fields = Field.createChain(size, signed, values, fields)
      
      for index, field in enumerate(fields):
         self.setParam(field, index == (len(fields) - 1))
   
   @property
   def size(self):
      return self.__format.getPacketSize(self.__buf) - self.__format.minPacketSize
   
   @property
   def rawBuffer(self):
      return self.__buf
   
   def _verifyCmdValidity(self, cmd):
      commandNumber = self.__format.getCommandNumber(self.__buf)
      
      if commandNumber != cmd:
         raise AssertionError(f"{self.__class__.__name__}: the command number must be {cmd}, not {commandNumber}.")
   
   def _verifySizeValidity(self, size):
      packetSize = self.__format.getPacketSize(self.__buf)
      
      if len(self.__buf) != packetSize:
         raise AssertionError(f"{self.__class__.__name__}: the internal buffer length ({len(self.__buf)}) isn't equal to the packet size stored in it ({packetSize}).")
      
      totalSize = size + self.__format.minPacketSize
      
      if packetSize != totalSize:
         raise AssertionError(f"{self.__class__.__name__}: the packet size must be {totalSize}, not {packetSize}.")



from .baseformat import BaseFormat

class SimpleFormat(BaseFormat):
   def __init__(self, marker, size, cmd, crc):
      super().__init__()
      
      if marker.size > 1:
         # see self.getPacketStartIndex() implementation.
         raise NotImplementedError()
      
      self._minPacketSize = marker.size + size.size + cmd.size + crc.size
      
      self.__marker = marker
      self.__size = size
      self.__cmd = cmd
      self._paramsOffset = cmd.nextOffset
      self.__crc = crc
   
   def getPacketStartIndex(self, buf, offset):
      return buf.index(self.__marker.value, offset)
   
   def hasEnoughBytes(self, buf, offset):
      return super().hasEnoughBytes(buf, offset) and (len(buf) - offset) >= self.getPacketSize(buf, offset)
   
   def getPacketSize(self, buf, offset = 0):
      return self._getField(buf, offset, self.__size).value
   
   def getCommandNumber(self, buf, offset = 0):
      return self._getField(buf, offset, self.__cmd).value
   
   def getParam(self, buf, param):
      return self._getField(buf, self._paramsOffset, param)
   
   def isValid(self, buf):
      if len(buf) < self.minPacketSize or self._getField(buf, 0, self.__marker).value != self.__marker.value:
         return False
      
      packetSize = self.getPacketSize(buf)
      self.__crc.offset = packetSize - self.__crc.size
      
      return packetSize >= self.minPacketSize   \
         and len(buf) == packetSize             \
         and self._getField(buf, 0, self.__crc).value == self._calcCrc(buf)
   
   def setCommandNumber(self, buf, commandNumber):
      self.__cmd.value = commandNumber
      
      self._setField(buf, 0, self.__cmd)
   
   def setParam(self, buf, param, finalize = False):
      self._setField(buf, self._paramsOffset, param)
      
      if finalize:
         self.finalizePacket(buf)
   
   def finalizePacket(self, buf):
      self._setField(buf, 0, self.__marker)
      
      self.__size.value = len(buf)
      self._setField(buf, 0, self.__size)
      
      self.__crc.offset = self.__size.value - self.__crc.size
      self.__crc.value = self._calcCrc(buf)
      self._setField(buf, 0, self.__crc)
   
   def _calcCrc(self, buf):
      crc = 0
      
      for i in range(self.__size.offset, len(buf) - self.__crc.size):
         crc += buf[i]
      
      return crc



from .packet import Packet

class Parser:
   class Handler:
      def __init__(self, packetType, handler):
         self.__packetType = packetType
         self.__handler = handler
      
      @property
      def packetType(self):
         return self.__packetType
      
      @property
      def handler(self):
         return self.__handler
   
   def __init__(self, format, defaultPacketType = Packet):
      self.__buf = bytearray()
      self.__handlers = {}
      self.__format = format
      self.__defaultPacketType = defaultPacketType
   
   def addHandler(self, handler):
      self.__handlers[handler.packetType.CMD] = handler
      
      return self
   
   def parse(self, data):
      packets = []
      
      self.__buf.extend(data)
      
      offset = 0
      
      while True:
         try:
            offset = self.__format.getPacketStartIndex(self.__buf, offset)
         
         except ValueError:
            offset = len(self.__buf)
            break
         
         if not self.__format.hasEnoughBytes(self.__buf, offset):
            break
         
         packetSize = self.__format.getPacketSize(self.__buf, offset)
         
         handler = self.__handlers.get(self.__format.getCommandNumber(self.__buf, offset))
         
         try:
            packet = (handler.packetType if handler else self.__defaultPacketType)(self.__format).wrap(self.__buf, start = offset, end = offset + packetSize)
         
         except ValueError:
            offset += 1
         
         else:
            if handler and handler.handler:
               handler.handler(packet)
               
            else:
               packets.append(packet)
            
            offset += packetSize
      
      del self.__buf[0:offset]
      
      return packets



from . import Formats
from .simpleformat import SimpleFormat
from ..field import Field

marker = Field(1, signed = False, value = 0x55)
size = Field(1, signed = False, previousField = marker)
cmd = Field(1, signed = False, previousField = size)

Formats.zx55v1 = SimpleFormat(marker, size, cmd, Field(2, signed = False))



from dataxmissionprotocol.version import __version__
from setuptools import find_packages, setup

with open("README.md", "r") as fh:
   long_description = fh.read()

setup(
   author="Ruben Shalimov",
   author_email="r_shalimov@inbox.ru",
   classifiers=[
      "Programming Language :: Python :: 3",
      "Operating System :: OS Independent"
   ],
   description="Data transmission protocol package",
   install_requires = [
      "simple-common-utils"
   ],
   long_description=long_description,
   long_description_content_type="text/markdown",
   name="dataxmissionprotocol",
   packages=find_packages(),
   url="https://github.com/RobinBobin/data-transmission-protocol",
   version=__version__
)



class BaseFormat:
   __byteorder = {
      None: "=",
      "little": "<",
      "big": ">"
   }
   
   def __init__(self):
      self.__byteorder = None
      self._minPacketSize = None
   
   @property
   def byteorder(self):
      return self.__byteorder
   
   @byteorder.setter
   def byteorder(self, byteorder):
      if byteorder not in BaseFormat.__byteorder.keys():
         raise ValueError() if isinstance(byteorder, str) else TypeError
      
      self.__byteorder = byteorder
   
   @property
   def minPacketSize(self):
      return self._minPacketSize
   
   def getPacketStartIndex(self, buf, offset):
      raise NotImplementedError()
   
   def hasEnoughBytes(self, buf, offset):
      return (len(buf) - offset) >= self._minPacketSize
   
   def getPacketSize(self, buf, offset = 0):
      raise NotImplementedError()
   
   def getCommandNumber(self, buf, offset = 0):
      raise NotImplementedError()
   
   def _getField(self, buf, offset, field):
      return field._get(buf, offset, BaseFormat.__byteorder[self.__byteorder])
   
   def _setField(self, buf, offset, field):
      field._set(buf, offset, BaseFormat.__byteorder[self.__byteorder])
